/**
 * \file flow-rs/src/node/mod.rs
 * MegFlow is Licensed under the Apache License, Version 2.0 (the "License")
 *
 * Copyright (c) 2019-2021 Megvii Inc. All rights reserved.
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT ARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 */
mod demux;
mod noop;
mod port;
mod reorder;
mod shared;
mod transform;

use crate::channel::ChannelStorage;
use crate::graph::Context;
use crate::graph::GraphSlice;
use crate::registry::Collect;
use crate::rt::task::JoinHandle;
use crate::{broker::BrokerClient, resource::ResourceCollection};
use anyhow::{anyhow, Result};
pub use port::*;
pub(crate) use shared::*;
use std::collections::BTreeSet;

#[doc(hidden)]
pub struct NodeSlice {
    pub cons: Box<dyn Fn(String, &toml::value::Table) -> Box<dyn Actor> + Send + Sync>,
    pub info: NodeInfo,
}
#[doc(hidden)]
#[derive(Clone)]
pub struct NodeInfo {
    pub inputs: Vec<String>,
    pub outputs: Vec<String>,
}
crate::collect!(String, NodeSlice);
/// Trait for interactiving with graph, which can be derived by `#[derive(Node)]`.
pub trait Node {
    fn set_port(&mut self, port_name: &str, tag: Option<u64>, channel: &ChannelStorage);
    fn set_port_dynamic(
        &mut self,
        local_key: u64,
        port_name: &str,
        target: String,
        cap: usize,
        broker: BrokerClient,
    );
    fn close(&mut self);
    fn is_allinp_closed(&self) -> bool;
}

/// Trait for interactiving with schedule, which can be derived by `#[derive(Actor)]`.
pub trait Actor: Node + Send + 'static {
    /// Run the actor
    fn start(self: Box<Self>, _: Context, _: ResourceCollection) -> JoinHandle<()> {
        unimplemented!()
    }
}

pub(crate) fn load(
    local_key: u64,
    config: &crate::config::interlayer::Node,
) -> Result<Vec<Box<dyn Actor>>> {
    if let Some(node) = NodeSlice::registry_local()
        .get(local_key)
        .get(&config.entity.ty)
    {
        Ok((0..config.cloned.unwrap_or(1))
            .into_iter()
            .map(|_| (node.cons)(config.entity.name.clone(), &config.entity.args))
            .collect())
    } else if let Some(graph) = GraphSlice::registry_local()
        .get(local_key)
        .get(&config.entity.ty)
    {
        (0..config.cloned.unwrap_or(1))
            .into_iter()
            .map(|_| (graph.cons)(config.entity.name.clone()))
            .map(|g| g.map(|g| Box::new(g) as Box<dyn Actor>))
            .collect()
    } else {
        Err(anyhow!(
            "unexpected node {}:{}",
            config.entity.name,
            config.entity.ty
        ))
    }
}

pub(crate) fn inputs(local_key: u64, ty: &str) -> Result<BTreeSet<String>> {
    if let Some(node) = NodeSlice::registry_local().get(local_key).get(ty) {
        Ok(node.info.inputs.iter().cloned().collect())
    } else if let Some(graph) = GraphSlice::registry_local().get(local_key).get(ty) {
        Ok(graph.info.inputs.iter().cloned().collect())
    } else {
        Err(anyhow!("unexpected node {}", ty))
    }
}

pub(crate) fn outputs(local_key: u64, ty: &str) -> Result<BTreeSet<String>> {
    if let Some(node) = NodeSlice::registry_local().get(local_key).get(ty) {
        Ok(node.info.outputs.iter().cloned().collect())
    } else if let Some(graph) = GraphSlice::registry_local().get(local_key).get(ty) {
        Ok(graph.info.outputs.iter().cloned().collect())
    } else {
        Err(anyhow!("unexpected node {}", ty))
    }
}
